import numpy as np
import scipy as sp
from skimage import measure
from skimage.util import view_as_windows


def get_offset(pattern):
    if pattern == 0:  # RGGB
        δy, δx = 0, 0
    elif pattern == 1:  # BGGR
        δy, δx = 1, 1
    elif pattern == 2:  # GRBG
        δy, δx = 0, 1
    elif pattern == 3:  # GBRG
        δy, δx = 1, 0
    else:
        raise ValueError(pattern)
    return δy, δx


def mosaic(img, δy, δx):
    img = img.copy()
    img[1 - δy::2, :, 0] = 0
    img[:, 1 - δx::2, 0] = 0
    img[δy::2, δx::2, 1] = 0
    img[1 - δy::2, 1 - δx::2, 1] = 0
    img[δy::2, :, 2] = 0
    img[:, δx::2, 2] = 0
    return img


def bilinear_demosaicing(mosaiced, δy, δx):
    r, g, b = mosaiced.transpose(2, 0, 1)
    fil_rb = np.array([[.25, .5, .25], [.5, 1, .5], [.25, .5, .25]])
    fil_g = np.array([[0, .25, 0], [.25, 1, .25], [0, .25, 0]])
    r = sp.signal.convolve2d(r, fil_rb, mode='valid')
    g = sp.signal.convolve2d(g, fil_g, mode='valid')
    b = sp.signal.convolve2d(b, fil_rb, mode='valid')
    demosaiced = np.stack((r, g, b)).transpose(1, 2, 0)
    return demosaiced


def get_residual(img):
    offsets = [get_offset(pattern) for pattern in range(4)]
    redemosaiced = np.stack([
        bilinear_demosaicing(mosaic(img, *offset), *offset)
        for offset in offsets
    ])
    residual = np.square(redemosaiced - img[None, 1:-1, 1:-1])
    return residual  # (4, Y, X, 3)


def create_sliding_windows(x, W, stride):
    """
    :param x: array-like, shape (4, Y, X, C)
    :return windows: np.ndarray, shape (n_windows, 4, W, W, C)
    :return cuts_y: np.ndarray, shape (n_windows,): Y coordinate of the centre of each window
    :return cuts_x: np.ndarray, shape (n_windows,)
    """
    _, Y, X, C = x.shape
    # windows = np.stack([
    # np.lib.stride_tricks.sliding_window_view(xx, (W, W),
    # axis=(0,
    # 1))[::stride, ::stride]
    # for xx in x
    # ])
    windows = view_as_windows(x, (4, W, W, 3), stride).reshape(-1, 4, W, W, 3)
    cuts_y = np.arange(W // 2, Y - W // 2 + 1)[::stride] - 0.5
    cuts_x = np.arange(W // 2, X - W // 2 + 1)[::stride] - 0.5
    # windows = windows.reshape(4, -1, C, W, W).transpose(1, 0, 3, 4, 2)
    return windows, cuts_y, cuts_x


def find_grid(x, multichannel_diag=False):
    """
    :param x: array-like, shape (n, 4, W, W, C)
    """
    n, _, W_Y, W_X, _ = x.shape
    averaged = np.mean(x, axis=(2, 3))  # (n, 4, 3)
    # residual_direct = np.sqrt(np.mean(averaged, axis=-1))  # (n, 4)
    # best_direct = np.argmin(residual_direct, axis=-1)

    if multichannel_diag:
        res_diag = np.sqrt(np.mean(averaged, axis=-1))
    else:
        res_diag = np.sqrt(averaged[:, :, 1])

    res_rb = np.sqrt(np.mean(averaged[:, :, [0, 1]], axis=-1))

    diff_diag = .5 * (res_diag[:, 0] + res_diag[:, 1] - res_diag[:, 2] -
                      res_diag[:, 3])
    best_diag = (diff_diag > 0).astype(int)  # 0 if XGGX, 1 if GXXG

    diff_rggb_bggr = res_rb[:, 0] - res_rb[:, 1]
    diff_grbg_gbrg = res_rb[:, 2] - res_rb[:, 3]

    diff_grid = best_diag * diff_grbg_gbrg + (1 - best_diag) * diff_rggb_bggr
    best_rb = (diff_grid > 0).astype(int)  # 0: RGGB or GRBG, 1:BGGR or GBRG
    best_grid = 2 * best_diag + best_rb  # 0: RGGB, 1: BGGR, 2: GRBG, 3: GBRG

    return best_grid, best_diag, diff_grid, diff_diag, diff_rggb_bggr, diff_grbg_gbrg


def find_forgeries(img, W=32, stride=4, multichannel_diag=False, threshold=2):
    Y, X, C = img.shape
    img = img.astype(float)
    img = img[:Y - Y % 2, :X - X % 2]
    residual = get_residual(img)
    windows, cuts_y, cuts_x = create_sliding_windows(residual, W, stride)
    cuts_y += 1
    cuts_x += 1

    def to_fullsize(data, method, fill=None):
        interpolator = sp.interpolate.RegularGridInterpolator(
            (cuts_y, cuts_x),
            data.reshape(len(cuts_y), len(cuts_x)),
            method=method,
            bounds_error=False,
            fill_value=fill)
        target_coords = np.asarray(
            np.meshgrid(np.arange(0, Y), np.arange(0, X)))
        result = interpolator(target_coords.reshape(2, -1).T).reshape(X, Y).T
        return result

    cY, cX = len(cuts_y), len(cuts_x)
    grid, diag, diff, diff_diag, diff_rggb_bggr, diff_grbg_gbrg = find_grid(
        windows, multichannel_diag=multichannel_diag)

    # diff_normalized = np.abs(diff) / np.max(np.abs(diff))
    # diff_diag_normalized = np.abs(diff_diag) / np.max(np.abs(diff_diag))

    main_grid, main_diag = find_grid(residual[None],
                                     multichannel_diag=multichannel_diag)[:2]
    main_grid = main_grid[0]
    main_diag = main_diag[0]

    diag[diff_diag == 0] = -1
    grid[diff == 0] = -1

    abs_diff = np.abs(diff)
    abs_diff_diag = np.abs(diff_diag)

    # connected maximum
    labels_diag, n_regions = measure.label(diag.reshape(cY, cX),
                                           background=main_diag,
                                           return_num=True)
    labels_diag = labels_diag.ravel()
    inconsistent_diag_con = np.zeros_like(diff_diag)
    for i_region in range(1, n_regions + 1):
        mask = labels_diag == i_region
        max_ = np.max(abs_diff_diag[mask])
        inconsistent_diag_con[mask] = max_

    labels_grid, n_regions = measure.label(grid.reshape(cY, cX),
                                           background=main_grid,
                                           return_num=True)
    labels_grid = labels_grid.ravel()
    inconsistent_grid_con = np.zeros_like(diff)
    for i_region in range(1, n_regions + 1):
        mask = labels_grid == i_region
        max_ = np.max(abs_diff[mask])
        inconsistent_grid_con[mask] = max_

    inconsistent_diag_raw = (diag != main_diag) * (diff_diag != 0)

    inconsistent_grid_raw = (grid != main_grid) * (diff != 0) * (diff_diag !=
                                                                 0)
    inconsistent_full_raw = np.maximum(inconsistent_diag_raw,
                                       inconsistent_grid_raw)

    inconsistent_full_con = np.maximum(inconsistent_diag_con,
                                       inconsistent_grid_con)

    inconsistent_grid = inconsistent_grid_con > threshold
    inconsistent_diag = inconsistent_diag_con > threshold
    inconsistent_full = inconsistent_full_con > threshold

    grid = to_fullsize(grid, 'nearest')
    diag = to_fullsize(diag, 'nearest')
    inconsistent_grid = to_fullsize(inconsistent_grid, 'nearest', 0)
    inconsistent_diag = to_fullsize(inconsistent_diag, 'nearest', 0)
    inconsistent_full = to_fullsize(inconsistent_full, 'nearest', 0)
    inconsistent_grid_con = to_fullsize(inconsistent_grid_con, 'linear', 0)
    inconsistent_diag_con = to_fullsize(inconsistent_diag_con, 'linear', 0)
    inconsistent_full_con = to_fullsize(inconsistent_full_con, 'linear', 0)
    inconsistent_grid_raw = to_fullsize(inconsistent_grid_raw, 'nearest', 0)
    inconsistent_diag_raw = to_fullsize(inconsistent_diag_raw, 'nearest', 0)
    inconsistent_full_raw = to_fullsize(inconsistent_full_raw, 'nearest', 0)
    diff = to_fullsize(diff, 'linear', 0)
    diff_rggb_bggr = to_fullsize(diff_rggb_bggr, 'linear', 0)
    diff_grbg_gbrg = to_fullsize(diff_grbg_gbrg, 'linear', 0)
    diff_diag = to_fullsize(diff_diag, 'linear', 0)

    return grid, diag, main_grid, inconsistent_grid, inconsistent_diag, inconsistent_full, inconsistent_grid_con, inconsistent_diag_con, inconsistent_full_con, inconsistent_grid_raw, inconsistent_diag_raw, inconsistent_full_raw, diff, diff_diag, diff_rggb_bggr, diff_grbg_gbrg
