#!/usr/bin/env python3
"""
Implementation of
Choi, C., Choi, J., & Lee, H. (2011). CFA pattern identification of digital cameras using intermediate value counting. MM&Sec '11.
"""

import argparse
import sys
from time import time

import imageio
import matplotlib as mpl
import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
from mpl_toolkits.axes_grid1 import make_axes_locatable

from doubledemosaick import find_forgeries
from postprocessing import (add_gaussian_noise, apply_median_filter,
                            gamma_correct, jpeg_compress)


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument("image", type=str)
    parser.add_argument("-o", "--out", default='out', type=str)
    parser.add_argument('-W',
                        '--window',
                        type=int,
                        default=32,
                        help="Sliding window size.")
    parser.add_argument(
        '-S',
        '--stride',
        type=int,
        default=None,
        help=
        "Stride of the window. Must be a multiple of two. Smaller values yield a finer map but are more costly."
    )

    diag_channel = parser.add_mutually_exclusive_group()
    diag_channel.add_argument('--green-diag',
                              action='store_false',
                              dest='multichannel_diag')
    diag_channel.add_argument('--multichannel-diag',
                              action='store_true',
                              dest='multichannel_diag')
    parser.set_defaults(multichannel_diag=True)

    parser.add_argument('--threshold',
                        default=2,
                        type=float,
                        help="Detection threshold")

    postprocessing = parser.add_argument_group("Postprocessing")
    postprocessing.add_argument("--noise", default=None, type=float)
    postprocessing.add_argument("--gray", action='store_true', default=False)
    postprocessing.add_argument("--hue", default=None, type=float)
    postprocessing.add_argument("--gamma", default=None, type=float)
    postprocessing.add_argument("--jpeg", default=None, type=int)
    postprocessing.add_argument("--median", action='store_true', default=False)
    return parser


def add_cbar(ax, plot):
    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', size='5%', pad='2%')
    cb = plt.colorbar(plot, cax=cax)
    cax.tick_params(labelsize=35)
    return cb


def process_image(img, args):
    img = img[:, :, :3]
    if args.gamma is not None:
        img = gamma_correct(img, args.gamma)
    if args.median:
        img = apply_median_filter(img)
    if args.noise is not None:
        img = add_gaussian_noise(img, args.noise, args.gray)
    if args.jpeg is not None:
        img = jpeg_compress(img, args.jpeg)
    return img


def main():
    t0 = time()
    parser = get_parser()
    args = parser.parse_args(sys.argv[1:])
    img = imageio.imread(args.image)
    assert img.ndim == 3
    img = process_image(img, args)
    out = args.out
    W = args.window
    stride = args.stride
    if stride is None:
        stride = W // 2 - W // 2 % 2
    t1 = time()
    grid, diag, main_grid, inconsistent_grid, inconsistent_diag, inconsistent_full, inconsistent_grid_con, inconsistent_diag_con, inconsistent_full_con, inconsistent_grid_raw, inconsistent_diag_raw, inconsistent_full_raw, diff, diff_diag, diff_rggb_bggr, diff_grbg_gbrg = find_forgeries(
        img,
        W=W,
        stride=stride,
        multichannel_diag=args.multichannel_diag,
        threshold=args.threshold)
    open(f"{out}_main_grid.txt", 'w').write(["RGGB", "BGGR", "GRBG",
                                             "GBRG"][main_grid])
    png_kwargs = {'optimize': False, 'compress_level': 7}
    t2 = time()
    # Plotting the results
    grid_cmap_colours = ['black', '#da8bc3', '#8c8c8c', '#8172b3', '#64b5cd']
    cmap_grid = ListedColormap(sns.color_palette(grid_cmap_colours).as_hex())
    cmap_diag = ListedColormap(
        sns.color_palette(grid_cmap_colours[:3]).as_hex())
    # Grids
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(grid, cmap=cmap_grid, vmin=-1, vmax=3)
    cb = add_cbar(ax, plot)
    cb.set_ticks([-.6, .2, 1, 1.8, 2.6])
    cb.set_ticklabels(['/', 'RGGB', 'BGGR', 'GRBG', 'GBRG'])
    ax.axis('off')
    plt.savefig(f"{out}_grids.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)
    # Diagonal grids
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(diag, cmap=cmap_diag, vmin=-1, vmax=1)
    cb = add_cbar(ax, plot)
    cb.set_ticks([-2 / 3, 0, 2 / 3])
    cb.set_ticklabels(['/', '*GG*', 'G**G'])
    ax.axis('off')
    plt.savefig(f"{out}_diag.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)

    # Soft forgeries, grid-level
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(inconsistent_grid_con,
                      cmap='mako',
                      vmin=0,
                      vmax=inconsistent_full_con.max())
    # cb = add_cbar(ax, plot)
    ax.axis('off')
    cb = add_cbar(ax, plot)
    plt.savefig(f"{out}_inconsistent_grid_con.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)

    # Soft forgeries, diagonally
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(inconsistent_diag_con,
                      cmap='mako',
                      vmin=0,
                      vmax=inconsistent_full_con.max())
    # cb = add_cbar(ax, plot)
    ax.axis('off')
    cb = add_cbar(ax, plot)
    plt.savefig(f"{out}_inconsistent_diag_con.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)

    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(inconsistent_full_con,
                      cmap='mako',
                      vmin=0,
                      vmax=inconsistent_full_con.max())
    # cb = add_cbar(ax, plot)
    ax.axis('off')
    cb = add_cbar(ax, plot)
    plt.savefig(f"{out}_inconsistent_full_con.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)

    norm = mpl.colors.CenteredNorm()
    cmap = 'icefire'
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(diff, norm=norm, cmap=cmap)
    cb = add_cbar(ax, plot)
    ax.axis('off')
    plt.savefig(f"{out}_diff_main.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)

    norm = mpl.colors.CenteredNorm()
    cmap = 'icefire'
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(diff_grbg_gbrg, norm=norm, cmap=cmap)
    cb = add_cbar(ax, plot)
    ax.axis('off')
    plt.savefig(f"{out}_diff_grbg_gbrg.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)

    norm = mpl.colors.CenteredNorm()
    cmap = 'icefire'
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(diff_rggb_bggr, norm=norm, cmap=cmap)
    cb = add_cbar(ax, plot)
    ax.axis('off')
    plt.savefig(f"{out}_diff_rggb_bggr.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)

    norm = mpl.colors.CenteredNorm()
    cmap = 'icefire'
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(diff_diag, norm=norm, cmap=cmap)
    cb = add_cbar(ax, plot)
    ax.axis('off')
    plt.savefig(f"{out}_diff_diag.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)
    t3 = time()
    print(
        f'Preprocessing done in {t1-t0:.1f}s.\nAlgorithm running time: {t2-t1:.1f}s.\nFigures saving time: {t3-t2:.1f}s.'
    )
    imageio.imsave(f"{out}_input.png", img, **png_kwargs)
    fig, ax = plt.subplots(figsize=(11, 11))
    plot = ax.matshow(inconsistent_full_con, cmap='mako', vmin=0)
    add_cbar(ax, plot)
    ax.axis('off')
    plt.savefig(f"{out}_inconsistent_full_con.png",
                bbox_inches='tight',
                pad_inches=0,
                transparent=False)
    imageio.imsave(f"{out}_inconsistent_grid.png",
                   (inconsistent_grid * 255).astype(np.uint8), **png_kwargs)
    imageio.imsave(f"{out}_inconsistent_diag.png",
                   (inconsistent_diag * 255).astype(np.uint8), **png_kwargs)
    imageio.imsave(f"{out}_inconsistent_grid_raw.png",
                   (inconsistent_grid_raw * 255).astype(np.uint8),
                   **png_kwargs)
    imageio.imsave(f"{out}_inconsistent_diag_raw.png",
                   (inconsistent_diag_raw * 255).astype(np.uint8),
                   **png_kwargs)
    imageio.imsave(f"{out}_inconsistent_full_raw.png",
                   (inconsistent_full_raw * 255).astype(np.uint8),
                   **png_kwargs)
    imageio.imsave(f"{out}_inconsistent_full.png",
                   (inconsistent_full * 255).astype(np.uint8), **png_kwargs)


if __name__ == "__main__":
    main()
